import abc
from typing import Any, Generic, Iterable, Optional, Union

from braandket.backend import Backend, ValuesType, get_default_backend
from braandket.space import BraSpace, HSpace, KetSpace, NumSpace, Space


class QTensor(Generic[ValuesType], abc.ABC):

    @classmethod
    def of(cls, values: Any, spaces: Iterable[Space], *, backend: Optional[Backend] = None) -> 'QTensor':
        if backend is None:
            backend = get_default_backend()
        spaces = tuple(spaces)

        try:
            from braandket import NumericTensor
            # noinspection PyTypeChecker
            return NumericTensor(values, spaces, backend)
        except TypeError:
            pass

        try:
            from .special import PureStateTensor
            # noinspection PyTypeChecker
            return PureStateTensor(values, spaces, backend)
        except TypeError:
            pass

        return QTensor(values, spaces, backend)

    def __init__(self, values: Any, spaces: Iterable[Space], backend: Backend):
        # check spaces
        spaces = tuple(spaces)
        for i in range(len(spaces)):
            for j in range(i + 1, len(spaces)):
                if spaces[i] == spaces[j]:
                    raise ValueError(f"Found duplicated spaces when constructing QTensor: {spaces[i]}")

        # check values
        values = backend.convert(values)
        values = backend.ensure_shape(values, tuple(space.n for space in spaces))

        # construct
        self._spaces = spaces
        self._values = values
        self._backend = backend

    @property
    def backend(self) -> Backend[ValuesType]:
        return self._backend

    def spawn(self, values: Any, spaces: Iterable[Space]) -> 'QTensor':
        """ create a new BackendTensor instance with the same backend """
        return QTensor.of(values, spaces, backend=self.backend)

    # basic operations

    @property
    def spaces(self) -> tuple[Space, ...]:
        return self._spaces

    def values(self, *slices: Union[Space, tuple[Space, Union[int, None]]]) -> ValuesType:
        """ get values from this tensor """
        return self.values_and_slices(*slices)[0]

    def values_and_slices(self,
            *slices: Union[Space, tuple[Space, Union[int, None]]]
    ) -> tuple[ValuesType, tuple[Union[Space, tuple[Space, int]], ...]]:
        if len(slices) == 0:
            return self._values, self._spaces

        # parse arguments
        sp_slices = slices
        spaces = []
        slices = []
        for sl in sp_slices:
            if isinstance(sl, Space):
                space, index = sl, None
            else:
                space, index = sl
            spaces.append(space)
            slices.append(index)

        # find axes of the specified spaces
        axes = []
        for space in spaces:
            axis = self._spaces.index(space)
            axes.append(axis)

        # fill axes of unspecified spaces
        axes_set = frozenset(axes)
        for axis, space in enumerate(self._spaces):
            if axis not in axes_set:
                axes.append(axis)
                spaces.append(space)
                slices.append(None)

        # values slices
        values_slices = tuple((slice(None) if sl is None else sl) for sl in slices)

        # manipulate values
        values = self._values
        values = self.backend.transpose(values, axes=axes)
        values = self.backend.slice(values, slices=values_slices)

        # slices for output
        sp_slices = tuple((sp if sl is None else (sp, sl)) for sp, sl in zip(spaces, slices))

        return values, sp_slices

    # common operations

    def __getitem__(self, items: Union[tuple[tuple[Space, int], ...], dict[Space, int]]) -> 'QTensor':
        if not isinstance(items, dict):
            items = dict(items)

        slices = tuple(items.get(sp, slice(None)) for sp in self._spaces)
        new_spaces = tuple(sp for sp in self._spaces if sp not in items)
        new_values = self.backend.slice(self._values, slices=slices)
        return self.spawn(new_values, new_spaces)

    def __iter__(self):
        raise TypeError("QTensor is not allowed to be iterated.")

    def __eq__(self, other):
        if self.is_scalar:
            return self.scalar() == other
        raise TypeError("The function __eq__() of QTensor not allowed, because it is ambiguous.")

    def __copy__(self) -> 'QTensor':
        return self.spawn(self.backend.copy(self._values), self._spaces)

    def copy(self) -> 'QTensor':
        return self.__copy__()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} spaces={repr(self._spaces)}, values={repr(self._values)}>"

    def __str__(self) -> str:
        return repr(self)

    # scalar operations

    @property
    def is_scalar(self) -> bool:
        return len(self.spaces) == 0

    def scalar(self) -> ValuesType:
        """ get a scalar value from zero-dimension tensor """
        if not self.is_scalar:
            raise ValueError(f"This QTensor is not a scalar. It has {len(self.spaces)} spaces.")
        return self.values()

    def __float__(self) -> float:
        return float(self.scalar())

    # linear operations

    def __pos__(self) -> 'QTensor':
        return self

    def __neg__(self) -> 'QTensor':
        return (-1) * self

    def __add__(self, other) -> 'QTensor':
        if not isinstance(other, QTensor):
            other = self.spawn(other, ())

        self_expanded, other_expanded = self._expand_tensors_for_addsub(self, other)

        new_spaces = tuple(self_expanded._spaces)
        self_values = self_expanded._values
        other_values = other_expanded.values(*new_spaces)
        new_values = self_expanded.backend.add(self_values, other_values)
        return self_expanded.spawn(new_values, new_spaces)

    def __radd__(self, other) -> 'QTensor':
        return self + other

    def __sub__(self, other) -> 'QTensor':
        return self + (-other)

    def __rsub__(self, other) -> 'QTensor':
        return -self + other

    @classmethod
    def _expand_tensors_for_addsub(cls, tensor0: 'QTensor', tensor1: 'QTensor') -> tuple['QTensor', 'QTensor']:
        from .operations import _broadcast_num_spaces
        tensor0, tensor1 = _broadcast_num_spaces(tensor0, tensor1)
        return tensor0, tensor1

    def __mul__(self, other) -> 'QTensor':
        if not isinstance(other, QTensor):
            other = self.spawn(other, ())

        self_expanded, other_expanded = self._expand_tensors_for_muldiv(self, other)

        new_spaces = self_expanded._spaces
        self_values = self_expanded._values
        other_values = other_expanded.values(*new_spaces)
        new_values = self_expanded.backend.mul(self_values, other_values)
        return self_expanded.spawn(new_values, new_spaces)

    def __rmul__(self, other) -> 'QTensor':
        return self * other

    def __truediv__(self, other) -> 'QTensor':
        if not isinstance(other, QTensor):
            other = self.spawn(other, ())

        self_expanded, other_expanded = self._expand_tensors_for_muldiv(self, other)

        new_spaces = self_expanded._spaces
        self_values = self_expanded._values
        other_values = other_expanded.values(*new_spaces)
        new_values = self_expanded.backend.div(self_values, other_values)
        return self_expanded.spawn(new_values, new_spaces)

    def __rtruediv__(self, other) -> 'QTensor':
        other = self.spawn(other, ())
        return other / self

    @classmethod
    def _expand_tensors_for_muldiv(cls, tensor0: 'QTensor', tensor1: 'QTensor') -> tuple['QTensor', 'QTensor']:
        from .operations import _broadcast_h_spaces, _broadcast_num_spaces
        tensor0, tensor1 = _broadcast_h_spaces(tensor0, tensor1)
        tensor0, tensor1 = _broadcast_num_spaces(tensor0, tensor1)
        return tensor0, tensor1

    # tensor operations

    def __matmul__(self, other) -> 'QTensor':
        if not isinstance(other, QTensor):
            return self * other

        self_expanded, other_expanded = self._expand_tensors_for_matmul(self, other)

        self_spaces = self_expanded._spaces
        self_values = self_expanded._values
        other_spaces = other_expanded._spaces
        other_values = other_expanded._values

        self_dot_axes, other_dot_axes = [], []
        self_num_axes, other_num_axes = [], []
        for self_axis, self_space in enumerate(self_spaces):
            if isinstance(self_space, KetSpace):
                pass  # ignored
            elif isinstance(self_space, BraSpace):
                try:
                    other_axis = other_spaces.index(self_space.ct)
                    self_dot_axes.append(self_axis)
                    other_dot_axes.append(other_axis)
                except ValueError:
                    pass  # ignored
            elif isinstance(self_space, NumSpace):
                other_axis = other_spaces.index(self_space)
                self_num_axes.append(self_axis)
                other_num_axes.append(other_axis)

        new_values, (self_rem_axes, other_rem_axes) = self_expanded.backend.dot(
            self_values, other_values,
            ndim0=len(self_spaces), ndim1=len(other_spaces),
            dot_axes=(self_dot_axes, other_dot_axes),
            bat_axes=(self_num_axes, other_num_axes))
        new_spaces = [
            *(self_spaces[axis] for axis in self_num_axes),
            *(self_spaces[axis] for axis in self_rem_axes),
            *(other_spaces[axis] for axis in other_rem_axes)]
        return self_expanded.spawn(new_values, new_spaces)

    def __rmatmul__(self, other) -> 'QTensor':
        return self @ other  # (other is scalar)

    @classmethod
    def _expand_tensors_for_matmul(cls, tensor0: 'QTensor', tensor1: 'QTensor') -> tuple['QTensor', 'QTensor']:
        from .operations import _broadcast_num_spaces
        # noinspection PyTypeChecker
        return _broadcast_num_spaces(tensor0, tensor1)

    # spaces operations

    @property
    def ct(self) -> 'QTensor':
        new_spaces = tuple((space.ct if isinstance(space, HSpace) else space) for space in self._spaces)
        new_values = self.backend.conj(self._values)
        return self.spawn(new_values, new_spaces)

    # specialize

    def as_numeric_tensor(self):
        from braandket import NumericTensor
        return NumericTensor.of(self)

    def as_pure_state_tensor(self):
        from .special import PureStateTensor
        return PureStateTensor.of(self)

    def as_mixed_state_tensor(self):
        from .special import MixedStateTensor
        return MixedStateTensor.of(self)

    def as_operator_tensor(self):
        from .special import OperatorTensor
        return OperatorTensor.of(self)

    # inflate & flatten

    @classmethod
    def inflate(cls,
            values: ValuesType,
            spaces: Iterable[Union[NumSpace, Iterable[KetSpace], Iterable[BraSpace]]], *,
            backend: Optional[Backend] = None
    ) -> 'QTensor':
        if backend is None:
            backend = get_default_backend()

        *num_spaces, ket_spaces, bra_spaces = spaces

        shape = (*(space.n for space in num_spaces),
                 *(space.n for space in ket_spaces),
                 *(space.n for space in bra_spaces))
        values = backend.reshape(values, shape)
        return cls.of(values, (*num_spaces, *ket_spaces, *bra_spaces), backend=backend)

    def flatten(self,
            spaces: Optional[Iterable[Union[NumSpace, KetSpace, BraSpace]]] = None, *,
            num_spaces: Optional[Iterable[NumSpace]] = None,
            ket_spaces: Optional[Iterable[KetSpace]] = None,
            bra_spaces: Optional[Iterable[BraSpace]] = None,
    ) -> tuple[ValuesType, tuple[Union[NumSpace, tuple[KetSpace, ...], tuple[BraSpace, ...]], ...]]:
        num_spaces_sl = tuple(space for space in self.spaces if isinstance(space, NumSpace))
        ket_spaces_sl = tuple(space for space in self.spaces if isinstance(space, KetSpace))
        bra_spaces_sl = tuple(space for space in self.spaces if isinstance(space, BraSpace))

        spaces = tuple(spaces) if spaces is not None else ()
        num_spaces_sp = tuple(space for space in spaces if isinstance(space, NumSpace))
        ket_spaces_sp = tuple(space for space in spaces if isinstance(space, KetSpace))
        bra_spaces_sp = tuple(space for space in spaces if isinstance(space, BraSpace))

        num_spaces = list(num_spaces) if num_spaces is not None else []
        for space in (*num_spaces_sp, *num_spaces_sl):
            if space not in num_spaces:
                num_spaces.append(space)

        ket_spaces = list(ket_spaces) if ket_spaces is not None else []
        for space in (*ket_spaces_sp, *ket_spaces_sl):
            if space not in ket_spaces:
                ket_spaces.append(space)

        bra_spaces_kt = tuple(space.ct for space in ket_spaces if space.ct in bra_spaces_sl)
        bra_spaces = list(bra_spaces) if bra_spaces is not None else []
        for space in (*bra_spaces_sp, *bra_spaces_kt, *bra_spaces_sl):
            if space not in bra_spaces:
                bra_spaces.append(space)

        shape = (*(space.n for space in num_spaces),
                 prod(*(space.n for space in ket_spaces)),
                 prod(*(space.n for space in bra_spaces)))
        values = self.backend.reshape(self.values(*num_spaces, *ket_spaces, *bra_spaces), shape)
        return values, (*num_spaces, tuple(ket_spaces), tuple(bra_spaces))

    def flattened_values(self,
            spaces: Optional[Iterable[Union[NumSpace, KetSpace, BraSpace]]] = None, *,
            num_spaces: Optional[Iterable[NumSpace]] = None,
            ket_spaces: Optional[Iterable[KetSpace]] = None,
            bra_spaces: Optional[Iterable[BraSpace]] = None,
    ) -> ValuesType:
        return self.flatten(
            spaces,
            num_spaces=num_spaces,
            ket_spaces=ket_spaces,
            bra_spaces=bra_spaces
        )[0]


# utils

def prod(*values: int) -> int:
    prod_value = 1
    for value in values:
        prod_value *= value
    return prod_value



from contextvars import ContextVar

from .backend import Backend
from .numpy_backend import numpy_backend

_backend_stack_context_var: ContextVar[tuple[Backend, ...]] = ContextVar("backend_stack", default=(numpy_backend,))


def get_default_backend() -> Backend:
    return _backend_stack_context_var.get()[-1]


def set_default_backend(backend: Backend):
    backend_stack = _backend_stack_context_var.get()
    _backend_stack_context_var.set((*backend_stack[:-1], backend))


def push_context_backend(backend: Backend):
    backend_stack = _backend_stack_context_var.get()
    _backend_stack_context_var.set((*backend_stack, backend))


def pop_context_backend() -> Backend:
    backend_stack = _backend_stack_context_var.get()
    if not len(backend_stack) > 1:
        raise RuntimeError("No enough backends in the context stack to pop!")
    _backend_stack_context_var.set(backend_stack[:-1])
    return backend_stack[-1]



from typing import Any, Optional

from braandket.backend import Backend
from .space import Space


class NumSpace(Space):
    def __init__(self, n: int, name: Optional[str] = None):
        self._n = n
        self._name = name

    # basics

    @property
    def n(self) -> int:
        return self._n

    @property
    def name(self) -> Optional[str]:
        return self._name

    # tensor constructors

    def zeros(self, *, backend: Optional[Backend] = None):
        from braandket.tensor import zeros
        return zeros(self, backend=backend)

    def ones(self, *, backend: Optional[Backend] = None):
        from braandket.tensor import ones
        return ones(self, backend=backend)

    def full(self, value: Any, *, backend: Optional[Backend] = None):
        return self.ones(backend=backend) * value

    def values(self, values: Any, *, backend: Optional[Backend] = None):
        from braandket.tensor import NumericTensor
        return NumericTensor.of(values, [self], backend=backend)



from .operations import abs, cos, exp, sin
from .operations import choose, take
from .operations import eigenstate, identity, ones, operator, projector, zeros
from .operations import one, pi, zero
from .operations import pow, sqrt, square
from .operations import prod, sum, sum_ct
from .special import MixedStateTensor, NumericTensor, OperatorTensor, PureStateTensor, StateTensor
from .tensor import QTensor



from typing import Any, Iterable, Optional, Union

import numpy as np

from .backend import Backend, ValuesType


class NumpyBackend(Backend[np.ndarray]):

    # basics

    def convert(self, values: Any, *, dtype=None) -> np.ndarray:
        return np.asarray(values, dtype=dtype)

    def copy(self, values: np.ndarray) -> np.ndarray:
        return np.copy(values)

    # constructors

    def zeros(self, shape: Iterable[int], *, dtype=np.float32) -> np.ndarray:
        return np.zeros(shape, dtype=dtype)

    def ones(self, shape: Iterable[int], *, dtype=np.float32) -> np.ndarray:
        return np.ones(shape, dtype=dtype)

    def onehot(self, index: int, size: int, *, dtype=np.float32) -> np.ndarray:
        values = np.zeros(size, dtype=dtype)
        values[index] = 1.0
        return values

    def eye(self, size: int, *, dtype=np.float32) -> np.ndarray:
        return np.eye(size, dtype=dtype)

    # unary operations

    def pow(self, values0: np.ndarray, values1: np.ndarray) -> np.ndarray:
        return np.power(values0, values1)

    def square(self, values: np.ndarray) -> np.ndarray:
        return np.square(values)

    def sqrt(self, values: np.ndarray) -> np.ndarray:
        return np.sqrt(values)

    def exp(self, values: np.ndarray) -> np.ndarray:
        return np.exp(values)

    def sin(self, values: np.ndarray) -> np.ndarray:
        return np.sin(values)

    def cos(self, values: np.ndarray) -> np.ndarray:
        return np.cos(values)

    def conj(self, values: np.ndarray) -> np.ndarray:
        return np.conj(values)

    def abs(self, values: np.ndarray) -> np.ndarray:
        return np.abs(values)

    # linear operations

    def add(self, values0: np.ndarray, values1: np.ndarray) -> np.ndarray:
        return values0 + values1

    def sub(self, values0: np.ndarray, values1: np.ndarray) -> np.ndarray:
        return values0 - values1

    def mul(self, values0: np.ndarray, values1: np.ndarray) -> np.ndarray:
        return values0 * values1

    def div(self, values0: np.ndarray, values1: np.ndarray) -> np.ndarray:
        return values0 / values1

    # tensor operations

    def ensure_shape(self, values: np.ndarray, shape: Iterable[int]):
        if np.shape(values) != tuple(shape):
            raise ValueError(f"Unexpected values shape! expected={shape}, actual={np.shape(values)}")
        return values

    def reshape(self, values: np.ndarray, shape: Iterable[int]) -> np.ndarray:
        return np.reshape(values, shape)

    def transpose(self, values: np.ndarray, *, axes: Iterable[int]) -> np.ndarray:
        return np.transpose(values, axes)

    def expand(self, values: np.ndarray, axes: Iterable[int], sizes: Optional[Iterable[int]] = None) -> np.ndarray:
        axes = tuple(axes)
        values = np.expand_dims(values, axes)
        if sizes is not None:
            sizes = tuple(sizes)
            for axis, size in zip(axes, sizes, strict=True):
                values = np.repeat(values, size, axis)
        return values

    def slice(self, values: np.ndarray, *, slices: Union[int, slice, Iterable[Union[int, slice]]]) -> np.ndarray:
        return values[slices]

    def trace(self, values: np.ndarray, axes: tuple[Iterable[int], Iterable[int]]) -> np.ndarray:
        axis_pairs = np.transpose(axes)  # [axes_n, 2]
        while len(axes) > 0:
            axis0, axis1 = axis_pairs[0]
            values = np.trace(values, axis0, axis1)
            axis_pairs = axis_pairs[1:]
            axis_pairs = np.where(axis_pairs > axis0, axis_pairs - 1, axis_pairs)
            axis_pairs = np.where(axis_pairs > axis1, axis_pairs - 1, axis_pairs)
        return values

    def diag(self, values: np.ndarray, axes: tuple[Iterable[int], Iterable[int]]) -> np.ndarray:
        axis_pairs = np.transpose(axes)  # [axes_n, 2]
        while len(axes) > 0:
            axis0, axis1 = axis_pairs[0]
            values = np.diagonal(values, axis0, axis1)
            axis_pairs = axis_pairs[1:]
            axis_pairs = np.where(axis_pairs > axis0, axis_pairs - 1, axis_pairs)
            axis_pairs = np.where(axis_pairs > axis1, axis_pairs - 1, axis_pairs)
        return values

    def dot(self,
            values0: np.ndarray, values1: np.ndarray, *,
            ndim0: int, ndim1: int,
            dot_axes: tuple[Iterable[int], Iterable[int]],
            bat_axes: tuple[Iterable[int], Iterable[int]],
    ) -> tuple[np.ndarray, tuple[tuple[int, ...], tuple[int, ...]]]:
        bat_axes0, bat_axes1 = tuple(bat_axes[0]), tuple(bat_axes[1])
        if not len(bat_axes0) == len(bat_axes1):
            raise ValueError("len(bat_axes[0]) != len(bat_axes[1])")
        del bat_axes

        dot_axes0, dot_axes1 = tuple(dot_axes[0]), tuple(dot_axes[1])
        if not len(dot_axes0) == len(dot_axes1):
            raise ValueError("len(dot_axes[0]) != len(dot_axes[1])")
        del dot_axes

        selected_axes0 = {*bat_axes0, *dot_axes0}
        if len(selected_axes0) != len(bat_axes0) + len(dot_axes0):
            raise ValueError("Found duplication for axes of values0 !")
        rem_axes0 = tuple(axis for axis in range(ndim0) if axis not in selected_axes0)

        selected_axes1 = {*bat_axes1, *dot_axes1}
        if len(selected_axes1) != len(bat_axes1) + len(dot_axes1):
            raise ValueError("Found duplication for axes of values1 !")
        rem_axes1 = tuple(axis for axis in range(ndim1) if axis not in selected_axes1)

        values0 = np.transpose(values0, [*bat_axes0, *rem_axes0, *dot_axes0])
        values1 = np.transpose(values1, [*bat_axes1, *rem_axes1, *dot_axes1])
        # [*bat_axes, *rem_axes, *dot_axes]

        bat_axes_n, dot_axes_n = len(bat_axes0), len(dot_axes0)
        rem_axes0_n, rem_axes1_n = len(rem_axes0), len(rem_axes1)

        values0 = np.expand_dims(values0, axis=tuple(bat_axes_n + rem_axes0_n + np.arange(rem_axes1_n)))
        values1 = np.expand_dims(values1, axis=tuple(bat_axes_n + np.arange(rem_axes0_n)))
        # [*bat_axes, *rem_axes0, *exp_axes0, *dot_axes]
        # [*bat_axes, *exp_axes1, *rem_axes1, *dot_axes]

        values = values0 * values1
        # [*bat_axes, *rem_axes0, *rem_axes1, *dot_axes]

        values = np.sum(values, axis=tuple(bat_axes_n + rem_axes0_n + rem_axes1_n + np.arange(dot_axes_n)))
        # [*bat_axes, *rem_axes0, *rem_axes1]

        return values, (rem_axes0, rem_axes1)

    # special

    def take(self, values: Iterable[ValuesType], indices: np.ndarray) -> np.ndarray:
        values = np.stack(values, axis=-1)
        indices = np.expand_dims(indices, axis=-1)
        values = np.take_along_axis(values, indices, axis=-1)
        values = np.squeeze(values, axis=-1)
        return values

    def choose(self, probs: Iterable[ValuesType]) -> tuple[np.ndarray, np.ndarray]:
        probs = np.stack(probs, axis=-1)  # [batch_size, choose_n]
        probs /= np.sum(probs, axis=-1, keepdims=True)
        probs = np.abs(probs)
        choice = np.apply_along_axis(lambda p: np.random.choice(len(p), p=p), axis=-1, arr=probs)  # [batch_size]
        return choice


numpy_backend = NumpyBackend()



import abc
import warnings
from typing import Optional

from braandket.backend import Backend
from .space import Space


class HSpace(Space, abc.ABC):
    @property
    @abc.abstractmethod
    def ct(self) -> 'HSpace':
        pass

    @property
    @abc.abstractmethod
    def ket(self) -> 'KetSpace':
        pass

    @property
    @abc.abstractmethod
    def bra(self) -> 'BraSpace':
        pass


class KetSpace(HSpace):
    def __init__(self, n: int, name: Optional[str] = None):
        super().__init__()
        self._n = n
        self._name = name

        self._bra = BraSpace(self)

    # basics

    @property
    def n(self) -> int:
        return self._n

    @property
    def name(self) -> Optional[str]:
        return self._name

    # spaces

    @property
    def ct(self) -> 'BraSpace':
        return self._bra

    @property
    def ket(self) -> 'KetSpace':
        return self

    @property
    def bra(self) -> 'BraSpace':
        return self._bra

    # tensor constructors

    def eigenstate(self, index: int, *, backend: Optional[Backend] = None):
        from braandket.tensor import eigenstate
        return eigenstate(self, index, backend=backend)

    def operator(self, ket_index: int, bra_index: int, *, backend: Optional[Backend] = None):
        from braandket.tensor import operator
        return operator(self, ket_index, bra_index, backend=backend)

    def projector(self, index: int, *, backend: Optional[Backend] = None):
        from braandket.tensor import projector
        return projector(self, index, backend=backend)

    def identity(self, *, backend: Optional[Backend] = None):
        from braandket.tensor import identity
        return identity(self, backend=backend)


class BraSpace(HSpace):
    def __new__(cls, ket: KetSpace):
        try:
            bra = ket.bra
            warnings.warn(
                "Please avoid creating a BraSpace yourself. "
                "Instead, use property \".ct\" on KetSpace instance.",
                category=UserWarning, stacklevel=2)
            return bra
        except AttributeError:
            return super().__new__(cls)

    def __init__(self, ket: KetSpace):
        super().__init__()

        if not isinstance(ket, KetSpace):
            raise TypeError("The parameter ket must be a KetSpace instance!")

        self._ket = ket

    # basics

    @property
    def n(self) -> int:
        return self._ket.n

    @property
    def name(self) -> Optional[str]:
        return self._ket.name

    # spaces

    @property
    def ct(self) -> KetSpace:
        return self._ket

    @property
    def ket(self) -> KetSpace:
        return self._ket

    @property
    def bra(self) -> 'BraSpace':
        return self



import abc
import weakref
from typing import Iterable, Optional

from braandket.backend import Backend
from braandket.tensor import PureStateTensor, StateTensor
from braandket.space.hilbert_space import KetSpace


# state

class QState:
    def __init__(self, tensor: StateTensor, related_systems: Iterable['QModel'] = ()):
        self._tensor = tensor
        self._related_systems = weakref.WeakSet(related_systems)
        for system in self._related_systems:
            # noinspection PyProtectedMember
            system._set_state(self)

    @property
    def tensor(self) -> StateTensor:
        return self._tensor

    @tensor.setter
    def tensor(self, tensor: StateTensor):
        self._tensor = tensor

    def _add_related_system(self, system: 'QModel'):
        self._related_systems.add(system)

    def __matmul__(self, other: 'QState') -> 'QState':
        if other is self:
            return self
        new_tensor = self._tensor @ other._tensor
        new_related_systems = (*self._related_systems, *other._related_systems)
        return QState(new_tensor, new_related_systems)

    @classmethod
    def prod(cls, *states: 'QState') -> 'QState':
        if len(states) == 0:
            return QState(PureStateTensor.of((), ()))
        if len(states) == 1:
            return states[0]
        return cls.prod(states[0] @ states[1], *states[2:])


# system

class QModel(abc.ABC):
    def __init__(self, *, name: Optional[str] = None):
        self._name = name
        self._state = None

    @property
    def name(self) -> Optional[str]:
        return self._name

    # particles

    @property
    @abc.abstractmethod
    def particles(self) -> frozenset['QParticle']:
        pass

    # state

    @property
    def state(self) -> QState:
        if self._state is None:
            self._state = self._init_state()
            # noinspection PyProtectedMember
            self._state._add_related_system(self)
        return self._state

    @abc.abstractmethod
    def _init_state(self) -> QState:
        pass

    def _set_state(self, state: QState):
        self._state = state

    # compose

    def __matmul__(self, other: 'QModel'):
        children = []

        if isinstance(self, QComposed) and self.name is None:
            children.extend(self)
        else:
            children.append(self)

        if isinstance(other, QComposed) and other.name is None:
            children.extend(other)
        else:
            children.append(other)

        return QComposed(*children)


class QParticle(QModel, KetSpace):
    def __init__(self, n: int, *, name: Optional[str] = None, backend: Optional[Backend] = None):
        QModel.__init__(self, name=name)
        KetSpace.__init__(self, n, name)
        self._backend = backend

    # particles

    @property
    def particles(self) -> frozenset['QParticle']:
        return frozenset((self,))

    # state

    def _init_state(self) -> QState:
        return QState(self.eigenstate(0, backend=self._backend))


class QComposed(QModel, Iterable[QModel]):
    def __init__(self, *children: QModel, name: Optional[str] = None):
        super().__init__(name=name)
        self._children = tuple(children)

    # children

    @property
    def children(self) -> tuple[QModel]:
        return self._children

    def __iter__(self):
        return iter(self._children)

    def __len__(self):
        return len(self._children)

    def __getitem__(self, item):
        return self._children[item]

    # particles

    @property
    def particles(self) -> frozenset['QParticle']:
        return frozenset.union(*(child.particles for child in self._children))

    # state

    def _init_state(self) -> QState:
        return QState.prod(*(system.state for system in self._children))



def iter_structure(structure):
    if structure is None:
        return
    try:
        structure_iter = iter(structure)
    except TypeError:
        structure_iter = None
    if structure_iter is not None:
        for structure_item in structure_iter:
            for sub_item in iter_structure(structure_item):
                yield sub_item
    else:
        yield structure


def map_structure(structure, map_func):
    if structure is None:
        return None
    if isinstance(structure, list):
        return [map_structure(item, map_func) for item in structure]
    elif isinstance(structure, tuple):
        return tuple(map_structure(item, map_func) for item in structure)
    else:
        return map_func(structure)



import abc
from typing import Optional


class Space(abc.ABC):

    # basics

    @property
    @abc.abstractmethod
    def n(self) -> int:
        pass

    @property
    @abc.abstractmethod
    def name(self) -> Optional[str]:
        pass

    # str & repr

    def __repr__(self):
        if self.name is None:
            return f"<{type(self).__name__}: n={self.n}>"
        else:
            return f"<{type(self).__name__}: n={self.n}, name={self.name}>"

    def __str__(self):
        return repr(self)



import abc
from typing import Any, Generic, Iterable, Optional, Union

from braandket.backend import Backend, ValuesType, get_default_backend
from braandket.space import KetSpace, NumSpace, Space
from .tensor import QTensor


# special tensors

class NumericTensor(QTensor[ValuesType]):

    @classmethod
    def of(cls,
            values: Union[QTensor, Any],
            spaces: Optional[Iterable[Space]] = None, *,
            backend: Optional[Backend] = None
    ) -> 'NumericTensor':
        if isinstance(values, NumericTensor):
            return values
        if isinstance(values, QTensor):
            # noinspection PyTypeChecker
            return cls(values._values, values._spaces, values._backend)
        if spaces is None:
            spaces = ()
        return cls(values, spaces, backend or get_default_backend())

    def __init__(self, values: Any, spaces: Iterable[Space], backend: Backend):
        spaces = tuple(spaces)
        for space in spaces:
            if not isinstance(space, NumSpace):
                raise TypeError(f"NumTensor accepts only NumSpace, got {space}!")
        super().__init__(values, spaces, backend)

    @property
    def spaces(self) -> tuple[NumSpace, ...]:
        # noinspection PyTypeChecker
        return super().spaces

    # linear operations

    def _mul_expanded_num_spaces(self, other: 'QTensor') -> 'NumericTensor':
        return self @ other


class StateTensor(QTensor[ValuesType], Generic[ValuesType], abc.ABC):
    @property
    @abc.abstractmethod
    def ket_spaces(self) -> tuple[KetSpace, ...]:
        pass

    # norm

    @abc.abstractmethod
    def norm(self) -> 'NumericTensor':
        pass

    @abc.abstractmethod
    def normalize(self) -> 'StateTensor':
        pass

    @abc.abstractmethod
    def norm_and_normalize(self) -> tuple['NumericTensor', 'StateTensor']:
        return self.norm(), self.normalize()

    # component

    @abc.abstractmethod
    def component(self, indices: Union[Iterable[tuple[KetSpace, int]], dict[KetSpace, int]]) -> 'StateTensor':
        pass

    @abc.abstractmethod
    def trace(self, *spaces: KetSpace) -> 'StateTensor':
        pass

    def remain(self, *spaces: KetSpace) -> 'StateTensor':
        all_ket_spaces = frozenset(self.ket_spaces)
        remain_ket_spaces_set = frozenset(spaces)
        traced_ket_spaces = all_ket_spaces - remain_ket_spaces_set
        return self.trace(*traced_ket_spaces)

    @abc.abstractmethod
    def probabilities(self,
            *spaces: Union[NumSpace, KetSpace, tuple[Union[NumSpace, KetSpace], Union[int, None]]]
    ) -> ValuesType:
        pass


class PureStateTensor(StateTensor[ValuesType]):

    @classmethod
    def of(cls,
            values: Union[QTensor, Any],
            spaces: Optional[Iterable[Union[NumSpace, KetSpace]]] = None, *,
            backend: Optional[Backend] = None
    ) -> 'PureStateTensor':
        if isinstance(values, PureStateTensor):
            return values
        if isinstance(values, QTensor):
            # noinspection PyTypeChecker
            return cls(values._values, values._spaces, values._backend)
        return cls(values, spaces, backend or get_default_backend())

    def __init__(self, values: Any, spaces: Iterable[Union[NumSpace, KetSpace]], backend: Backend):
        spaces = tuple(spaces)
        for space in spaces:
            if not isinstance(space, (NumSpace, KetSpace)):
                raise TypeError(f"PureStateTensor accepts only KetSpace and NumSpace, got {space}!")
        super().__init__(values, spaces, backend)
        self._ket_spaces = tuple(space for space in spaces if isinstance(space, KetSpace))

    @property
    def spaces(self) -> tuple[Union[NumSpace, KetSpace], ...]:
        # noinspection PyTypeChecker
        return super().spaces

    @property
    def ket_spaces(self) -> tuple[KetSpace, ...]:
        return self._ket_spaces

    # norm

    def norm(self) -> 'NumericTensor':
        from .operations import abs
        return abs(NumericTensor.of(self.ct @ self))

    def normalize(self) -> 'PureStateTensor':
        return self.norm_and_normalize()[1]

    def norm_and_normalize(self) -> tuple['NumericTensor', 'PureStateTensor']:
        from .operations import sqrt
        norm = self.norm()
        normalized = self / sqrt(norm)
        return norm, normalized

    # component

    def component(self, indices: Union[Iterable[tuple[KetSpace, int]], dict[KetSpace, int]]) -> 'PureStateTensor':
        if isinstance(indices, dict):
            indices = tuple(indices.items())
        else:
            indices = tuple(indices)
        return PureStateTensor.of(self[indices])

    def trace(self, *spaces: KetSpace) -> 'MixedStateTensor':
        return MixedStateTensor.of(self @ self.ct).trace(*spaces)

    def amplitudes(self,
            *spaces: Union[NumSpace, KetSpace, tuple[Union[NumSpace, KetSpace], Union[int, None]]]
    ) -> ValuesType:
        return self.values(*spaces)

    def probabilities(self,
            *spaces: Union[NumSpace, KetSpace, tuple[Union[NumSpace, KetSpace], Union[int, None]]]
    ) -> ValuesType:
        amplitudes = self.amplitudes(*spaces)
        return self.backend.mul(self.backend.conj(amplitudes), amplitudes)


class MixedStateTensor(StateTensor[ValuesType]):
    @classmethod
    def of(cls,
            values: Union[QTensor, Any],
            spaces: Optional[Iterable[Union[NumSpace, KetSpace]]] = None, *,
            backend: Optional[Backend] = None
    ) -> 'MixedStateTensor':
        if isinstance(values, MixedStateTensor):
            return values
        if isinstance(values, QTensor):
            # noinspection PyTypeChecker
            return cls(values._values, values._spaces, values._backend)
        return cls(values, spaces, backend or get_default_backend())

    def __init__(self, values: Any, spaces: Iterable[Space], backend: Backend):
        spaces = tuple(spaces)
        ket_spaces = _match_spaces_pairs(spaces)
        super().__init__(values, spaces, backend)
        self._ket_spaces = ket_spaces

    @property
    def ket_spaces(self) -> tuple[KetSpace, ...]:
        return self._ket_spaces

    # norm

    def norm(self) -> 'NumericTensor':
        return NumericTensor.of(self.trace(*self.ket_spaces))

    def normalize(self) -> 'MixedStateTensor':
        return self.norm_and_normalize()[1]

    def norm_and_normalize(self) -> tuple['NumericTensor', 'MixedStateTensor']:
        norm = self.norm()
        normalized = self / norm
        return norm, normalized

    # component

    def component(self, indices: Union[Iterable[tuple[KetSpace, int]], dict[KetSpace, int]]) -> 'MixedStateTensor':
        if isinstance(indices, dict):
            ket_indices = tuple(indices.items())
        else:
            ket_indices = tuple(indices)
        bra_indices = tuple((ket_space.ct, index) for ket_space, index in ket_indices)
        return MixedStateTensor.of(self[ket_indices + bra_indices])

    def trace(self, *spaces: KetSpace) -> 'MixedStateTensor':
        values, spaces = self.values_and_slices(*spaces)
        ket_axes, bra_axes = _index_spaces_pairs(spaces)
        return self.backend.trace(values, (ket_axes, bra_axes))

    def probabilities(self,
            *spaces: Union[NumSpace, KetSpace, tuple[Union[NumSpace, KetSpace], Union[int, None]]]
    ) -> ValuesType:
        bra_spaces = []
        for space_or_pair in spaces:
            if isinstance(space_or_pair, Space):
                space, index = space_or_pair, None
            else:
                space, index = space_or_pair
            if isinstance(space, KetSpace):
                if index is None:
                    bra_spaces.append(space.ct)
                else:
                    bra_spaces.append((space.ct, index))
        return self.values(*spaces, *bra_spaces)


class OperatorTensor(QTensor[ValuesType]):
    @classmethod
    def of(cls,
            values: Union[QTensor, Any],
            spaces: Optional[Iterable[Space]] = None, *,
            backend: Optional[Backend] = None
    ) -> 'OperatorTensor':
        if isinstance(values, OperatorTensor):
            return values
        if isinstance(values, QTensor):
            # noinspection PyTypeChecker
            return cls(values._values, values._spaces, values._backend)
        return cls(values, spaces, backend or get_default_backend())

    @classmethod
    def from_matrix(cls,
            matrix: ValuesType,
            ket_spaces: Iterable[KetSpace],
            num_spaces: Iterable[KetSpace] = (), *,
            backend: Optional[Backend] = None
    ) -> 'OperatorTensor':
        num_spaces = tuple(num_spaces)
        ket_spaces = tuple(ket_spaces)
        bra_spaces = tuple(space.ct for space in ket_spaces)
        spaces = num_spaces + ket_spaces + bra_spaces
        backend = backend or get_default_backend()
        shape = tuple(space.n for space in spaces)
        values = backend.reshape(matrix, shape)
        return cls(values, spaces, backend)

    def __init__(self, values: Any, spaces: Iterable[Space], backend: Backend):
        spaces = tuple(spaces)
        ket_spaces = _match_spaces_pairs(spaces)
        super().__init__(values, spaces, backend)
        self._ket_spaces = ket_spaces

    @property
    def ket_spaces(self) -> tuple[KetSpace, ...]:
        return self._ket_spaces

    # linear operations

    def __add__(self, other) -> 'OperatorTensor':
        return OperatorTensor.of(super().__add__(other))

    def __mul__(self, other) -> 'OperatorTensor':
        return OperatorTensor.of(super().__mul__(other))

    @classmethod
    def _expand_tensors_for_addsub(cls, tensor0: 'QTensor', tensor1: 'QTensor') -> tuple['QTensor', 'QTensor']:
        assert isinstance(tensor0, OperatorTensor)
        assert isinstance(tensor1, OperatorTensor)
        tensor0_ket_spaces = tensor0.ket_spaces
        tensor1_ket_spaces = tensor1.ket_spaces

        from .operations import _expand_with_identities
        tensor0 = _expand_with_identities(tensor0, *tensor1_ket_spaces)
        tensor1 = _expand_with_identities(tensor1, *tensor0_ket_spaces)
        return super()._expand_tensors_for_addsub(tensor0, tensor1)

    # spaces operations

    @property
    def ct(self) -> 'OperatorTensor':
        return OperatorTensor.of(super().ct)

    def __matmul__(self, other) -> 'QTensor':
        tensor = super().__matmul__(other)
        if isinstance(other, OperatorTensor):
            tensor = OperatorTensor.of(tensor)
        return tensor


# utils

def _index_spaces_pairs(spaces: Iterable[Space]) -> tuple[tuple[int, ...], tuple[int, ...]]:
    spaces = tuple(spaces)
    ket_axes = []
    bra_axes = []
    for axis, space in enumerate(spaces):
        if isinstance(space, KetSpace):
            ket_axes.append(axis)
            bra_axes.append(spaces.index(space.ct))
    return tuple(ket_axes), tuple(bra_axes)


def _match_spaces_pairs(spaces: Iterable[Space]) -> tuple[KetSpace, ...]:
    spaces = tuple(spaces)
    ket_axes, bra_axes = _index_spaces_pairs(spaces)
    return tuple(spaces[axis] for axis in ket_axes)



import math
from typing import Any, Callable, Iterable, Optional, Union

from braandket.backend import Backend, ValuesType, get_default_backend
from braandket.space import HSpace, KetSpace, NumSpace, Space
from .special import NumericTensor, OperatorTensor, PureStateTensor
from .tensor import QTensor


# constants

def zero(*, backend: Optional[Backend] = None):
    return QTensor.of(0, (), backend=backend)


def one(*, backend: Optional[Backend] = None):
    return QTensor.of(1, (), backend=backend)


pi = math.pi


# constructors

def zeros(space: NumSpace, *, backend: Optional[Backend] = None) -> NumericTensor:
    backend = backend or get_default_backend()
    values = backend.zeros((space.n,))
    return NumericTensor(values, (space,), backend)


def ones(space: NumSpace, *, backend: Optional[Backend] = None) -> NumericTensor:
    backend = backend or get_default_backend()
    values = backend.ones((space.n,))
    return NumericTensor(values, (space,), backend)


def eigenstate(space: KetSpace, index: int, *, backend: Optional[Backend] = None) -> PureStateTensor:
    backend = backend or get_default_backend()
    values = backend.onehot(index, space.n)
    return PureStateTensor(values, (space,), backend)


def operator(space: KetSpace, ket_index: int, bra_index: int, *, backend: Optional[Backend] = None) -> OperatorTensor:
    ket_vector = eigenstate(space, ket_index, backend=backend)
    bra_vector = eigenstate(space, bra_index, backend=backend).ct
    return OperatorTensor.of(ket_vector @ bra_vector)


def projector(space: KetSpace, index: int, *, backend: Optional[Backend] = None) -> OperatorTensor:
    return operator(space, index, index, backend=backend)


def identity(space: KetSpace, *, backend: Optional[Backend] = None) -> OperatorTensor:
    backend = backend or get_default_backend()
    values = backend.eye(space.n)
    return OperatorTensor(values, (space, space.ct), backend)


# prod & sum

def prod(*items: QTensor, backend: Optional[Backend] = None) -> QTensor:
    if len(items) == 0:
        return one(backend=backend)
    x = items[0]
    for item in items[1:]:
        x = x @ item
    return x


def sum(*items: QTensor, backend: Optional[Backend] = None) -> QTensor:
    if len(items) == 0:
        return zero(backend=backend)
    x = items[0]
    for item in items[1:]:
        x = x + item
    return x


def sum_ct(*items, backend: Optional[Backend] = None) -> QTensor:
    s = sum(*items, backend=backend)
    return s + s.ct


# broadcast

def _broadcast(tensor: QTensor, *spaces: Space) -> QTensor:
    # extract non-existed spaces
    spaces = tuple(space for space in spaces if space not in tensor.spaces)

    # shortcut if there is no non-existed spaces
    if len(spaces) == 0:
        return tensor

    # construct new tensor
    new_spaces = (*spaces, *tensor._spaces)
    new_values = tensor.backend.expand(tensor._values, axes=range(len(spaces)), sizes=(space.n for space in spaces))
    return tensor.spawn(new_values, new_spaces)


def _broadcast_num_spaces(tensor0: QTensor, tensor1: QTensor, *tensors: QTensor) -> tuple[QTensor, ...]:
    if len(tensors) == 0:
        # extract num_spaces
        tensor0_num_spaces = tuple(space for space in tensor0.spaces if isinstance(space, NumSpace))
        tensor1_num_spaces = tuple(space for space in tensor1.spaces if isinstance(space, NumSpace))

        # broadcast num_spaces
        tensor0 = _broadcast(tensor0, *tensor1_num_spaces)
        tensor1 = _broadcast(tensor1, *tensor0_num_spaces)

        return tensor0, tensor1
    else:
        tensors = [tensor0, tensor1, *tensors]
        for i in range(len(tensors)):
            for j in range(i + 1, len(tensors)):
                tensors[i], tensors[j] = _broadcast_num_spaces(tensors[i], tensors[j])
        return tuple(tensors)


def _broadcast_h_spaces(tensor0: QTensor, tensor1: QTensor) -> tuple[QTensor, QTensor]:
    # extract h_spaces
    tensor0_h_spaces = tuple(space for space in tensor0.spaces if isinstance(space, HSpace))
    tensor1_h_spaces = tuple(space for space in tensor1.spaces if isinstance(space, HSpace))

    # broadcast h_spaces
    if tensor0_h_spaces != tensor1_h_spaces:
        if len(tensor0_h_spaces) == 0:
            tensor0 = _broadcast(tensor0, *tensor1_h_spaces)
        elif len(tensor1_h_spaces) == 0:
            tensor1 = _broadcast(tensor1, *tensor0_h_spaces)
        else:
            raise ValueError("Can not broadcast tensors with different set of h_spaces set.")

    return tensor0, tensor1


def _expand_with_identities(tensor: OperatorTensor, *spaces: KetSpace):
    return prod(tensor, *(space.identity(backend=tensor.backend) for space in spaces))


# numeric

def _construct_wrapped_unary_op(func: Callable[[ValuesType], ValuesType]):
    def wrapped_op(value: Union[NumericTensor, Any]) -> NumericTensor:
        value = NumericTensor.of(value)
        op = getattr(value.backend, func.__name__)
        # noinspection PyProtectedMember
        return value.spawn(op(value._values), value._spaces)

    return wrapped_op


pow = _construct_wrapped_unary_op(Backend.pow)
square = _construct_wrapped_unary_op(Backend.square)
sqrt = _construct_wrapped_unary_op(Backend.sqrt)
exp = _construct_wrapped_unary_op(Backend.exp)
sin = _construct_wrapped_unary_op(Backend.sin)
cos = _construct_wrapped_unary_op(Backend.cos)
abs = _construct_wrapped_unary_op(Backend.abs)


# choose

def choose(probs: Iterable[Union[NumericTensor, Any]]) -> NumericTensor:
    probs = tuple(NumericTensor.of(prob) for prob in probs)
    if len(probs) == 0:
        raise ValueError("Can not choose from empty set.")

    probs = _broadcast_num_spaces(*probs)
    spaces = tuple(probs[0].spaces)
    backend = probs[0].backend

    probs_values = tuple(prob.values(*spaces) for prob in probs)
    choice_values = backend.choose(probs_values)
    choice = NumericTensor.of(choice_values, spaces, backend=backend)
    return choice


def take(values: Iterable[QTensor], indices: NumericTensor) -> QTensor:
    values = _broadcast_num_spaces(*values)
    # TODO automatically expand for OperatorTensor
    assert len(values) > 0
    spaces = tuple(values[0].spaces)
    backend = values[0].backend
    for value in values:
        indices, _ = _broadcast_h_spaces(indices, value)

    values_values = tuple(value.values(*spaces) for value in values)
    indices_values = indices.values(*spaces)
    taken_values = backend.take(values_values, indices_values)
    taken = NumericTensor.of(taken_values, spaces, backend=backend)
    return taken



import abc
from typing import Any, Generic, Iterable, Optional, Union
from typing import TypeVar

ValuesType = TypeVar('ValuesType')


class Backend(Generic[ValuesType], abc.ABC):

    def __enter__(self):
        from .default import push_context_backend
        push_context_backend(self)

    def __exit__(self, exc_type, exc_val, exc_tb):
        from .default import pop_context_backend
        pop_context_backend()

    # basics

    @abc.abstractmethod
    def convert(self, values: Any) -> ValuesType:
        pass

    @abc.abstractmethod
    def copy(self, values: ValuesType) -> ValuesType:
        pass

    # constructors

    @abc.abstractmethod
    def zeros(self, shape: Iterable[int]) -> ValuesType:
        pass

    @abc.abstractmethod
    def ones(self, shape: Iterable[int]) -> ValuesType:
        pass

    @abc.abstractmethod
    def onehot(self, index: int, size: int) -> ValuesType:
        pass

    @abc.abstractmethod
    def eye(self, size: int) -> ValuesType:
        pass

    # linear operations

    @abc.abstractmethod
    def add(self, values0: ValuesType, values1: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def sub(self, values0: ValuesType, values1: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def mul(self, values0: ValuesType, values1: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def div(self, values0: ValuesType, values1: ValuesType) -> ValuesType:
        pass

    # non-linear operations

    @abc.abstractmethod
    def pow(self, values0: ValuesType, values1: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def square(self, values: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def sqrt(self, values: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def exp(self, values: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def sin(self, values: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def cos(self, values: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def conj(self, values: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def abs(self, values: ValuesType) -> ValuesType:
        pass

    # tensor operations

    @abc.abstractmethod
    def ensure_shape(self, values: ValuesType, shape: Iterable[int]) -> ValuesType:
        pass

    @abc.abstractmethod
    def reshape(self, values: ValuesType, shape: Iterable[int]) -> ValuesType:
        pass

    @abc.abstractmethod
    def transpose(self, values: ValuesType, *, axes: Iterable[int]) -> ValuesType:
        pass

    @abc.abstractmethod
    def expand(self, values: ValuesType, axes: Iterable[int], sizes: Optional[Iterable[int]] = None) -> ValuesType:
        pass

    @abc.abstractmethod
    def slice(self, values: ValuesType, *, slices: Union[int, slice, Iterable[Union[int, slice]]]) -> ValuesType:
        pass

    @abc.abstractmethod
    def trace(self, values: ValuesType, axes: tuple[Iterable[int], Iterable[int]]) -> ValuesType:
        pass

    @abc.abstractmethod
    def diag(self, values: ValuesType, axes: tuple[Iterable[int], Iterable[int]]) -> ValuesType:
        pass

    @abc.abstractmethod
    def dot(self,
            values0: ValuesType, values1: ValuesType, *,
            ndim0: int, ndim1: int,
            dot_axes: tuple[Iterable[int], Iterable[int]],
            bat_axes: tuple[Iterable[int], Iterable[int]],
    ) -> tuple[ValuesType, tuple[tuple[int, ...], tuple[int, ...]]]:
        pass

    # special

    @abc.abstractmethod
    def take(self, values: Iterable[ValuesType], indices: ValuesType) -> ValuesType:
        pass

    @abc.abstractmethod
    def choose(self, probs: Iterable[ValuesType]) -> ValuesType:
        pass



from typing import Any, Iterable, Optional, Union

import tensorflow as tf

from .backend import Backend


class TensorflowBackend(Backend[tf.Tensor]):

    # basics

    def convert(self, values: Any, *, dtype=None) -> tf.Tensor:
        values = tf.convert_to_tensor(values)
        if dtype is not None:
            values = tf.cast(values, dtype=dtype)
        elif values.dtype in (tf.int8, tf.int16, tf.int32, tf.int64):
            values = tf.cast(values, dtype=tf.int32)
        elif values.dtype in (tf.float16, tf.float32, tf.float64):
            values = tf.cast(values, dtype=tf.float32)
        elif values.dtype in (tf.complex64, tf.complex128):
            values = tf.cast(values, dtype=tf.complex64)
        else:
            raise TypeError(f"Unsupported dtype of values: {values.dtype}")
        return values

    def copy(self, values: tf.Tensor) -> tf.Tensor:
        return values

    def _auto_cast(self, values0: tf.Tensor, values1: tf.Tensor) -> tuple[tf.Tensor, tf.Tensor]:
        ordered_dtypes = [tf.complex64, tf.float32, tf.int32]
        dt0 = ordered_dtypes.index(values0.dtype)
        dt1 = ordered_dtypes.index(values1.dtype)
        dtype = ordered_dtypes[dt0] if dt0 < dt1 else ordered_dtypes[dt1]
        values0 = tf.cast(values0, dtype=dtype)
        values1 = tf.cast(values1, dtype=dtype)
        return values0, values1

    # constructors

    def zeros(self, shape: Iterable[int], *, dtype=tf.complex64) -> tf.Tensor:
        return tf.zeros(shape, dtype=dtype)

    def ones(self, shape: Iterable[int], *, dtype=tf.complex64) -> tf.Tensor:
        return tf.ones(shape, dtype=dtype)

    def onehot(self, index: int, size: int, *, dtype=tf.complex64) -> tf.Tensor:
        one_value = tf.ones((), dtype=dtype)
        zero_value = tf.zeros((), dtype=dtype)
        return tf.one_hot(index, size, on_value=one_value, off_value=zero_value)

    def eye(self, size: int, *, dtype=tf.complex64) -> tf.Tensor:
        return tf.eye(size, dtype=dtype)

    # unary operations

    def pow(self, values0: tf.Tensor, values1: tf.Tensor) -> tf.Tensor:
        return tf.pow(values0, values1)

    def square(self, values: tf.Tensor) -> tf.Tensor:
        return tf.square(values)

    def sqrt(self, values: tf.Tensor) -> tf.Tensor:
        return tf.sqrt(values)

    def exp(self, values: tf.Tensor) -> tf.Tensor:
        return tf.exp(values)

    def sin(self, values: tf.Tensor) -> tf.Tensor:
        return tf.sin(values)

    def cos(self, values: tf.Tensor) -> tf.Tensor:
        return tf.cos(values)

    def conj(self, values: tf.Tensor) -> tf.Tensor:
        return tf.math.conj(values)

    def abs(self, values: tf.Tensor) -> tf.Tensor:
        return tf.abs(values)

    # linear operations

    def add(self, values0: tf.Tensor, values1: tf.Tensor) -> tf.Tensor:
        values0, values1 = self._auto_cast(values0, values1)
        return values0 + values1

    def sub(self, values0: tf.Tensor, values1: tf.Tensor) -> tf.Tensor:
        values0, values1 = self._auto_cast(values0, values1)
        return values0 - values1

    def mul(self, values0: tf.Tensor, values1: tf.Tensor) -> tf.Tensor:
        values0, values1 = self._auto_cast(values0, values1)
        return values0 * values1

    def div(self, values0: tf.Tensor, values1: tf.Tensor) -> tf.Tensor:
        values0, values1 = self._auto_cast(values0, values1)
        return values0 / values1

    # operator operations

    def ensure_shape(self, values: tf.Tensor, shape: Iterable[int]) -> tf.Tensor:
        return tf.ensure_shape(values, shape)

    def reshape(self, values: tf.Tensor, shape: Iterable[int]) -> tf.Tensor:
        return tf.reshape(values, shape)

    def transpose(self, values: tf.Tensor, *, axes: Iterable[int]) -> tf.Tensor:
        return tf.transpose(values, axes)

    def expand(self, values: tf.Tensor, axes: Iterable[int], sizes: Optional[Iterable[int]] = None) -> tf.Tensor:
        for axis in axes:
            values = tf.expand_dims(values, axis)
        if sizes is not None:
            sizes = tuple(sizes)
            for axis, size in zip(axes, sizes, strict=True):
                values = tf.repeat(values, size, axis)
        return values

    def slice(self, values: tf.Tensor, *, slices: Union[int, slice, Iterable[Union[int, slice]]]) -> tf.Tensor:
        return values[slices]

    def trace(self, values: tf.Tensor, axes: tuple[Iterable[int], Iterable[int]]) -> tf.Tensor:
        axis_pairs = tf.transpose(axes)  # [axes_n, 2]
        while len(axes) > 0:
            axis0, axis1 = axis_pairs[0]
            values = tf.linalg.trace(values, axis0, axis1)
            axis_pairs = axis_pairs[1:]
            axis_pairs = tf.where(axis_pairs > axis0, axis_pairs - 1, axis_pairs)
            axis_pairs = tf.where(axis_pairs > axis1, axis_pairs - 1, axis_pairs)
        return values

    def diag(self, values: tf.Tensor, axes: tuple[Iterable[int], Iterable[int]]) -> tf.Tensor:
        axis_pairs = tf.transpose(axes)  # [axes_n, 2]
        while len(axes) > 0:
            axis0, axis1 = axis_pairs[0]
            values = tf.linalg.diag(values, axis0, axis1)
            axis_pairs = axis_pairs[1:]
            axis_pairs = tf.where(axis_pairs > axis0, axis_pairs - 1, axis_pairs)
            axis_pairs = tf.where(axis_pairs > axis1, axis_pairs - 1, axis_pairs)
        return values

    def dot(self,
            values0: tf.Tensor, values1: tf.Tensor, *,
            ndim0: int, ndim1: int,
            dot_axes: tuple[Iterable[int], Iterable[int]],
            bat_axes: tuple[Iterable[int], Iterable[int]],
    ) -> tuple[tf.Tensor, tuple[tuple[int, ...], tuple[int, ...]]]:
        bat_axes0, bat_axes1 = tuple(bat_axes[0]), tuple(bat_axes[1])
        if not len(bat_axes0) == len(bat_axes1):
            raise ValueError("len(bat_axes[0]) != len(bat_axes[1])")
        del bat_axes

        dot_axes0, dot_axes1 = tuple(dot_axes[0]), tuple(dot_axes[1])
        if not len(dot_axes0) == len(dot_axes1):
            raise ValueError("len(dot_axes[0]) != len(dot_axes[1])")
        del dot_axes

        selected_axes0 = {*bat_axes0, *dot_axes0}
        if len(selected_axes0) != len(bat_axes0) + len(dot_axes0):
            raise ValueError("Found duplication for axes of values0 !")
        rem_axes0 = tuple(axis for axis in range(ndim0) if axis not in selected_axes0)

        selected_axes1 = {*bat_axes1, *dot_axes1}
        if len(selected_axes1) != len(bat_axes1) + len(dot_axes1):
            raise ValueError("Found duplication for axes of values1 !")
        rem_axes1 = tuple(axis for axis in range(ndim1) if axis not in selected_axes1)

        values0 = tf.transpose(values0, [*bat_axes0, *rem_axes0, *dot_axes0])
        values1 = tf.transpose(values1, [*bat_axes1, *rem_axes1, *dot_axes1])

        bat_axes_n, dot_axes_n = len(bat_axes0), len(dot_axes0)
        rem_axes0_n, rem_axes1_n = len(rem_axes0), len(rem_axes1)

        values0 = self.expand(values0, tuple((i + bat_axes_n + rem_axes0_n) for i in range(rem_axes1_n)))
        values1 = self.expand(values1, tuple((i + bat_axes_n) for i in range(rem_axes0_n)))
        # [*bat_axes, *rem_axes0, *exp_axes0, *dot_axes]
        # [*bat_axes, *exp_axes1, *rem_axes1, *dot_axes]

        values = self.mul(values0, values1)
        # [*bat_axes, *rem_axes0, *rem_axes1, *dot_axes]

        values = tf.reduce_sum(values, tuple((i + bat_axes_n + rem_axes0_n + rem_axes1_n) for i in range(dot_axes_n)))
        # [*bat_axes, *rem_axes0, *rem_axes1]

        return values, (rem_axes0, rem_axes1)

    # special

    def take(self, values: Iterable[tf.Tensor], indices: tf.Tensor) -> tf.Tensor:
        values = tf.stack(values, axis=-1)
        indices = tf.expand_dims(indices, axis=-1)
        indices = tf.cast(indices, tf.int32)
        values = tf.experimental.numpy.take_along_axis(values, indices, axis=-1)
        values = tf.squeeze(values, axis=-1)
        return values

    def choose(self, probs: Iterable[tf.Tensor]) -> tf.Tensor:
        probs = tf.stack(probs, axis=-1)  # [*batch_shape, choose_n]
        batch_shape = tf.shape(probs)[:-1]
        batch_size = tf.reduce_prod(batch_shape)
        probs = tf.reshape(probs, [batch_size, -1])  # [batch_size, choose_n]
        logits = tf.math.abs(tf.math.log(probs))

        choice = tf.random.categorical(logits, 1)  # [batch_size, 1]
        choice = choice[:, 0]  # [batch_size]
        choice = tf.reshape(choice, batch_shape)  # [*batch_shape]
        return choice


tensorflow_backend = TensorflowBackend()



from .backend import Backend, ValuesType
from .default import get_default_backend, set_default_backend

try:
    from .numpy_backend import NumpyBackend, numpy_backend
except ImportError:
    pass

try:
    from .tensorflow_backend import TensorflowBackend, tensorflow_backend
except ImportError:
    pass
